from .example_module import heyo
import argparse
from . import __setup_gr

DESCRIPTION = """
The 'go-over' provide a set of tools to process input from different sources,
and generate JSON output that is easy to be process with Liquid in Jekyll.

It was build to help me maintaing my blog at `https://josealobato.com`.
"""

parser = argparse.ArgumentParser(description=DESCRIPTION,
    formatter_class=argparse.RawDescriptionHelpFormatter)

subparsers = parser.add_subparsers(dest="command", help = "Available Commands for go-over.")

# Global artument options
parser.add_argument("-v", "--verbose", action="store_true", help="Print extra information.")

# Configure subparers.
# Every new parsert represents a different subtool. It will have its
# configuration in a module in at `go_over` lever. 
__setup_gr.setup_gr_parser(subparsers)

# Function to build the global options to be intected in all tools.
def _build_options(args):
    options = { 
        "verbose": args.verbose,
        "rewrite_complementary_data": args.force_complementary_rewrite
    }
    return options

# Script entry point defined in the setup.cfg file.
# It will be called after initializing the script, so 
# when the reset of the code in this file is run.
def cli_processor():

    # Parsing the CLI arguments.
    args = parser.parse_args()
    # Creating the options dictionary to be injected to all tools.
    options = _build_options(args)
    # Check if any command has been used... 
    if args.command:
        # if so, call it (notice that every parser is defined with a method to call).
        args.func(args, options) 
    else:
        # otherwise show the help
        parser.print_help()


""" 
Bookshelf Module
A book shelf represent a collection of books.
It offers action that could be performed on such collection.
"""
import itertools

from datetime import datetime
from typing import Dict, List
import logging

# from .bookread import BookRead
from .book import Book, BookRead
from .loader import Loader

# pylint: disable=C0301
# line too long

class Bookshelf:
    """ A collection of books. """

    def __init__(self) -> None:
        self.books = []

    @property
    def sorted_books(self) -> List['Book']:
        """ Return the books sorted by date."""
        return sorted(self.books, reverse=True,
                    key=lambda b: b.date_read if b.date_read is not None else datetime.min)

    @property
    def all_reads(self) -> List['BookRead']:
        """ Get a list of all reads """
        all_reads = [b.reads() for b in self.books]
        # flatten the list
        return list(itertools.chain(*all_reads))

    @property
    def tags(self) -> List:
        """ Getting a list of the existing tags """
        tags_collection = []
        for book in self.books:
            for tag in book.tags:
                if tag not in tags_collection:
                    tags_collection.append(tag)
        return tags_collection

    @property
    def books_by_tags(self) -> Dict:
        """ Create a dictionary that will hold and entry for very tag with an array of books """
        tags_collection = {}
        for book in self.books:
            if not book.read:
                continue
            for tag in book.tags:
                tags_collection.setdefault(tag, [])
                tags_collection[tag].append(book.dictionary)
        return tags_collection

    @property
    def years_with_books(self) -> List[int]:
        """ Get the list of years that has books """
        years = set()
        for book in self.books:
            if book.date_read:
                years.add(book.date_read.year)
        return list(years)

    @property
    def statistics(self) -> Dict:
        """ Dictionary with the bookshelf statistics """
        stats = {}
        years = self.years_with_books
        for year in years:
            stats[str(year)] = self.statistics_for_year(year)
        return stats

    def reads_on_year(self, year: int) -> List[BookRead]:
        """ Return the list of reads ordered by year in invers order. """
        reads = [read for read in self.all_reads if read.date.year == year]
        return sorted(reads, reverse=True, key=lambda b: b.date)

    def read_on_unkown_date(self) -> List[Book]:
        """ Read books but with unknonw date """
        books = [book for book in self.books if book.read_in_unknown_date]
        return books

    def reading_books(self) -> List[Book]:
        """ Getting the list of currently reading books """
        books = [book for book in self.books if book.currently_reading]
        return books

    def favourites(self) -> List[Book]:
        """ Get the books marked as favourites """
        books = [book for book in self.books if book.is_favourite]
        return books

    def to_read(self) -> List[Book]:
        """ Get the books marked as to read sorted by position """
        books = [book for book in self.books if book.is_to_read]
        return sorted(books, key=lambda x: x.to_read_position)
    
    def statistics_for_year(self, year: int) -> Dict:
        """ Getting the statistics for a year """
        stats = { "read": 0 }
        languages = {}
        book_that_year = [b for b in self.books if b.read_on_year(year)]
        stats["read"] = len(book_that_year)
        for book in book_that_year:
            # Language
            language = book.language
            if language:
                languages.setdefault(language, 0)
                languages[language] += 1
        stats["languages"] = languages
        return stats

    @classmethod
    def build_from_loader(cls, loader: Loader) -> 'Bookshelf':
        """ load content from the give"""
        log = logging.getLogger(__name__)
        log.info(f"Loading goodreads csv data from file: {loader.source_path}")
        shelf = Bookshelf()
        loaded_data = loader.load()
        for dictionary in loaded_data:
            book = Book.build_from_dictionary(dictionary)
            shelf.books.append(book)
        log.info(f"A total of {len(shelf.books)} loaded.")
        log.info("Goodreads csv data loaded.")
        return shelf

    def update_with_complementary_information(self, books_complementary_data: List) -> None:
        """  Getting the complementary data it will update the bookshelf with the new data. """
        # books_comp_data = complementary_data["books"]
        for book in self.books:
            # More  info:
            # https://vald-phoenix.github.io/pylint-errors/plerr/errors/variables/W0640.html
            # pylint: disable=W0640
            equal_ids = lambda b: b["id"] == book.identifier
            comp_data = list(filter(equal_ids, books_complementary_data))
            if comp_data:
                book.update_with_complentary_data_dictionary(comp_data[0])



import os
import json
from typing import Dict, List
import logging

_FILE_EXTENSION = ".json"

class Printer:
    """ Printer for a concrete file.
        It needs the full path of a file to print on."""

    def __init__(self, target_path: str) -> None:
        """ Build a printer with the full path (path + filename + extension)."""
        self.__target_path = target_path
        self.__logger = logging.getLogger(__name__)

    @classmethod
    def in_current_directory(cls, file_name: str) -> 'Printer':
        """ Return a printer to print on a given filename within the current folder."""
        current_path = os.getcwd()
        full_path = os.path.join(current_path, file_name + _FILE_EXTENSION)
        return cls(full_path)

    @classmethod
    def in_directory(cls, directory_path: str, file_name: str) -> 'Printer':
        """ Returns a printer to print in a given directory with a given file name."""
        full_path = os.path.join(directory_path, file_name + _FILE_EXTENSION)
        return cls(full_path)

    @property
    def target_path_exist(self) -> bool:
        return os.path.exists(self.__target_path)

    def _create_folder_if_needed(self):
        """ Create the target folder if needed. """
        dir_name = os.path.dirname(self.__target_path)
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

    def dump(self, dict: Dict) -> None:
        """ Getting a dictionary and a file name it will create a file in given path 
        with the content in JSON. """
        self.__logger.info(f"Write file at: {self.__target_path}")
        self._create_folder_if_needed()
        with open(self.__target_path, 'w', encoding="utf-8") as json_file:
            json.dump(dict, json_file, indent=4)

class OnDirectoryPrinter:
    """ Printer for files in a directory
        It needs the full path of a directory to print files on."""

    def __init__(self, directory_path: str="./") -> None:
        """ Build a printer with the full directory path. If no path is given it assumes the current directory. """
        self.__directory_path = directory_path
        self.__logger = logging.getLogger(__name__)

    @property
    def directory_path_exist(self) -> bool:
        return os.path.exists(self.__directory_path)

    @property
    def printer_path(self) -> str:
        return self.__directory_path

    def _create_directory_if_needed(self):
        """ Create the target folder if needed. """
        dir_name = os.path.abspath(self.__directory_path)
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

    def dump(self, dict: Dict, file_name: str) -> None:
        """ Getting a dictionary and a file name it will create a file in given path 
        with the content in JSON. """
        full_path = os.path.join(self.__directory_path, file_name + _FILE_EXTENSION)
        self.__logger.info(f"Write file at: {full_path}")
        self._create_directory_if_needed()
        with open(full_path, 'w', encoding="utf-8") as json_file:
            json.dump(dict, json_file, indent=4)


from typing import Dict, List
from .bookshelf import Bookshelf
from .loader import Loader
from .printer import Printer
import logging

class BookshelfComplement:
    """ The bookshelf complement manages the complementary data of the bookshelf.
        This is data that comes from other sources (manually created) instead of Goodreads,
        to complement its information."""

    def __init__(self, loader: Loader) -> None:
        self.__data_dictionary = None
        self.logger = logging.getLogger(__name__)
        self.loader = loader
        self.printer = self.loader.printer()

    @property
    def complementary_information_list(self) -> List:
        """ Getting a list of the complementary information for books. """
        return self.__data_dictionary['books']

    def update_from_shelf(self, shelf: 'Bookshelf') -> None:
        """ With the object create with the minimum data (the source path) it will load it,
            update it from the existing shelf and prepare it for its use. Finally it will update
            the source complementary file with the available data."""
        self.__generate_template_if_needed(shelf)
        self.__load()
        self.__update_source_from_bookshelf(shelf)

    def __update_source_from_bookshelf(self, shelf: 'Bookshelf') -> None:
        """ Will update the source file with The books on the shelf. """
        data_dictionary = {"books": []}
        for book in shelf.books:
            data_dictionary['books'].append(book.complementary_data_dictionary)
        self.printer.dump(data_dictionary)

    def update_from_shelf_if_needed(self, shelf: 'Bookshelf') -> None:
        """ With the object create with the minimum data (the source path) it will load it
            update it from the existing shelf and prepare it for its use. Finally it will update
            the source complementary file with the available data for new books only."""
        self.__generate_template_if_needed(shelf)
        self.__load()
        self.__update_source_from_bookshelf_if_needed(shelf)

    def __update_source_from_bookshelf_if_needed(self, shelf: 'Bookshelf') -> None:
        """ Will update the source file with new books if any. """
        is_dirty = False
        all_complementary_data_ids = [c["id"] for c in self.complementary_information_list]
        books_not_in_complementary = [b for b in shelf.books
                                    if b.identifier not in all_complementary_data_ids]
        if books_not_in_complementary:
            is_dirty = True
        for book in books_not_in_complementary:
            self.__data_dictionary['books'].append(book.complementary_data_dictionary)
            self.logger.info(f'Append new book to complementary data with id {book.identifier}')

        if is_dirty:
            self.printer.dump(self.__data_dictionary)

    def __load(self) -> None:
        """ It will load the JSON file into an ivar """
        self.__data_dictionary = self.loader.load()

    def __generate_template_if_needed(self, shelf: 'Bookshelf') -> None:
        """ If the complement data file does not exist it will generate it from the given shelf.
            It will generate a default entry for every book."""
        if not self.loader.source_exist:
            self.logger.info(f"Generate complementary data file: {self.loader.source_path}")
            books_complementary_data = [b.complementary_data_dictionary for b in shelf.books]
            data_dictionary = { 'books': books_complementary_data }
            self.printer.dump(data_dictionary)


# GOODREADS Method and Parser 
import argparse
from .goodreads import processor

DEFAULT_GR_CVS_FILE_PATH = "goodreads_library_export.csv"
DEFAULT_GR_JSON_COMPLEMENT_FILE_PATH = "books_complement.json"
DEFAULT_GR_RESULTS_PATH = "./results"

# Action when the Goodreads parser is selected
def gr_processor(args, options):
    processor.process(args.gr_csv_path, args.complement_json_path, args.results_path, options)

# Create Parser for Good Reads processing
GR_DESCRIPTION = """
This script convert the goodreads csv to a nicer format for jekyll reading.

It takes the `cvs` generated from GoodReads (My Books > Import Export) and 
another JSON file with complementary data, and generates JSON files prepared 
to be read with Liquid and displayed on the Jekyll blog.

The outpu JSON files are:
* reading.
* one for every year.
* one with books with no date.

The idea is that those JSON files cound be maintained by hand if GoodReads stops working.

NOTES on the complementary file.
* It contains fields that are missing on GR. 
* It matches with the GR data using the GR book Id.
* If this file is not there it will be generated from the existing books. You will need to fill it
  and run it again.
* If a new book is added, during processing this file will be updated. You will need to add
  the information and run it again.
"""
def setup_gr_parser(subparsers):
    parser_gr = subparsers.add_parser("goodreads", 
        aliases=['gr'],
        help="Process a Goodreads CSV (see help).",
        description=GR_DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser_gr.add_argument("-g", "--gr_csv_path", 
        default=DEFAULT_GR_CVS_FILE_PATH,
        help="CVS exported from goodreads.")
    parser_gr.add_argument("-c", "--complement_json_path", 
        default=DEFAULT_GR_JSON_COMPLEMENT_FILE_PATH,
        help="JSON file with data to complement goodreads data.")
    parser_gr.add_argument("-r", "--results_path",
        default=DEFAULT_GR_RESULTS_PATH,
        help="Path were the output will be written.")
    parser_gr.add_argument("-f", "--force_complementary_rewrite", 
        action="store_true", 
        help="Force the rewrite of the complementary file (no data loss).")
    parser_gr.set_defaults(func=gr_processor) # Important: this is the function to call if the parser is selected.



import logging
from typing import Dict, List

from go_over.goodreads.generator import Generator

from .loader import Loader
from .printer import OnDirectoryPrinter
from .bookshelf import Bookshelf
from .bookshelfComplement import BookshelfComplement

def process(gr_csv_path: str, complement_json_path: str, results_path: str, options: Dict): 

    options.setdefault("verbose", False)
    if options["verbose"]:
        logging.basicConfig(level=logging.INFO, format='%(message)s')

    logger = logging.getLogger(__name__)
    
    # Load.
    logger.info("\n * Load source CVS:")
    loader = Loader(gr_csv_path)
    bookshelf = Bookshelf.build_from_loader(loader)
    
    # Load complementary data and update it if needed with the existing shelf.
    # Note that if complementray data file is not there, it will be created.
    # Existing information on the complementary data wont be overriden.
    logger.info("\n * Load and update complementary json:")
    complement_loader = Loader(complement_json_path)
    bookshelf_complement = BookshelfComplement(complement_loader)
    bookshelf_complement.update_from_shelf_if_needed(bookshelf)

    # Update Shell with complementary data.
    # Now with the complete complementary data update the shelf.
    # This will apply all changes that the user made to the complementary data.
    logger.info("\n * Update shelf with complementary data.")
    bookshelf.update_with_complementary_information(
        bookshelf_complement.complementary_information_list
    )

    # If requested force the regeneration of the complementary data with the
    # shelf. Notice that that new complementary data will update the non existing 
    # complementary data parts with the information from the books.
    options.setdefault("rewrite_complementary_data", False)
    if options["rewrite_complementary_data"]:
        logger.info("\n * Force write the complementary data.")
        bookshelf_complement.update_from_shelf(bookshelf)

    # Print out all the wanted results from the new shelf.
    logger.info("\n * Print out resulting files:")
    printer = OnDirectoryPrinter(results_path)
    generator = Generator(bookshelf, printer)
    generator.execute()

    logger.info("\n * Done!")



""" 
    Book Module 
    Notice that it contains the BookRead class as well.
    That was done to avoid circular imports. They both depend on each other.
"""
from datetime import datetime
from typing import Dict, List
import logging
import re

def to_read_position_from_string(string: str) -> int:
    pattern = r"to-read \(\#(\d+)\)"
    matched = re.findall(pattern, string)
    if matched:
        value = matched[0]
        return int(value)
    else:
        return 0

# pylint: disable=R0902
class Book:
    """
    This class represents a book.
    """

    def __init__(self, **kwargs) -> None:
        self.identifier = kwargs['Book Id']
        self.title = kwargs['Title']
        self.author = kwargs['Author']
        self.currently_reading = kwargs["Exclusive Shelf"] == "currently-reading"
        self.read = kwargs["Exclusive Shelf"] == "read"
        date_read = kwargs["Date Read"]
        if date_read == "":
            self.date_read = None
            self.read_dates = []
        else:
            self.date_read = datetime.strptime(date_read, "%Y/%m/%d")
            self.read_dates = [self.date_read]
        self.rating = int(kwargs['My Rating'])
        self.language = "EN"
        self.__tags = "nonfiction"
        self.document_format = "audiobook"
        self.my_review_url = None
        self.is_favourite = False
        self.is_to_read = kwargs["Exclusive Shelf"] == "to-read"
        if "Bookshelves with positions" in kwargs:
            self.to_read_position = to_read_position_from_string(kwargs["Bookshelves with positions"])
        else:
            self.to_read_position = 0

    def __str__(self):
        return f'"{self.title}" by {self.author} read on {self.date_read}'

    @classmethod
    def build_from_dictionary(cls, complementary_data: Dict) -> 'Book':
        """ Build a book from a dictionary of the needed data """
        return cls(**complementary_data)

    def update_with_complentary_data_dictionary(self, complementary_data: Dict) -> None:
        """ Update the book with the data comming from a dictionary with complementary data.
            Notice that the complementary data many miss keys. """
        complementary_data.setdefault("title", self.title)
        self.title = complementary_data["title"]
        
        complementary_data.setdefault("language", self.language)
        self.language = complementary_data["language"]

        complementary_data.setdefault("tags", self.__tags)
        self.__tags = complementary_data["tags"]

        complementary_data.setdefault("format", self.document_format)
        self.document_format = complementary_data["format"]

        complementary_data.setdefault("my_review_url")
        self.my_review_url = complementary_data["my_review_url"]

        complementary_data.setdefault("read_dates", [])
        new_read_dates = complementary_data["read_dates"]
        if new_read_dates:
            if self.read:
                self.add_reads(new_read_dates)
            else:
                print(f"[Warning] trying to add a read date to unread book with id {self.identifier}.")
        
        complementary_data.setdefault("is_favourite", False)
        self.is_favourite = complementary_data["is_favourite"]
        if self.is_favourite:
            logger = logging.getLogger(__name__)
            logger.info(f"- Favourite book: {self.title[:30]}...")

    @property
    def last_read(self) -> datetime:
        """ Return the last time the book was read or None"""
        if not self.read_dates:
            return None
        sorted_dates = sorted(self.read_dates, reverse=True)
        return sorted_dates[0]

    def add_reads(self, new_reads: List) -> None:
        """ Add the new reas with format %Y/%m/%d"""
        new_dates = [datetime.strptime(new_read, "%Y/%m/%d") for new_read in new_reads]
        all_read_dates = set(self.read_dates + new_dates)
        self.read_dates = list(all_read_dates)

    def read_on_year(self, year: int) -> bool:
        """ Tell if the book was read in a given year."""
        reads_on_that_year = [read for read in self.read_dates if read.year == year]
        return len(reads_on_that_year) > 0

    @property
    def read_years(self) -> List[int]:
        """ Get the list of year in which this book was read """
        years = set()
        for date in self.read_dates:
            years.add(date.year)
        return list(years)

    @property
    def read_in_unknown_date(self) -> bool:
        """ Tell is the book was read but do not know when. """
        return self.last_read == None and self.read

    @property
    def tags(self) -> List:
        """ Getting a list of tags. """
        return list(map(str.strip, self.__tags.split(",")))

    @property
    def dictionary(self) -> Dict:
        """ Create a dictionary with the Book ready to export. """
        result = {}
        result['id'] = self.identifier
        result['author'] = self.author
        result['title'] = self.title
        result['rating'] = self.rating
        # Note, this date format is the one preffered by Liquid.
        result['date'] = self.date_read.strftime("%Y-%m-%d") if self.date_read else None
        result['tags'] = self.tags
        result['format'] = self.document_format
        result['goodreads_url'] = "https://www.goodreads.com/book/show/" + self.identifier
        result['language'] = self.language
        result['my_review_url'] = self.my_review_url
        result['is_favourite'] = self.is_favourite
        result['to_read_position'] = self.to_read_position
        return result

    @property
    def complementary_data_dictionary(self) -> Dict:
        """ From a book it created a dictionary with only the complementary data """
        result = {}
        result['id'] = self.identifier
        result['title'] = self.title
        result['language'] = self.language
        result['tags'] = self.__tags
        result['format'] = self.document_format
        result['my_review_url'] = self.my_review_url
        if self.read_dates:
            result['read_dates'] = [d.strftime("%Y/%m/%d") for d in self.read_dates]
        else:
            result['read_dates'] = None 
        result['is_favourite'] = self.is_favourite
            
        return result

    def reads(self) -> List['BookRead']:
        """ Getting the list of current reads of this book """
        reads = []
        sorted_dates = sorted(self.read_dates)
        for i, read_date in enumerate(sorted_dates):
            new_read = BookRead(self, read_date, i + 1)
            reads.append(new_read)
        return reads


class BookRead:
    """
    This class represents the read of a book.
    That means that it contains a single data and the information about 
    the read numbe (it was the first time read, the second, etc).
    """

    def __init__(self, book: Book, read_date: datetime, read_number: int = 1) -> None:
        self.__book = book
        self.__date = read_date
        self.__read_number = read_number

    @property
    def date(self) -> datetime:
        """ Get the read date """
        return self.__date

    @property
    def read_number(self) -> int:
        """ Get the read number """
        return self.__read_number

    @property
    def identifier(self) -> str:
        """ Get the book id """
        return self.__book.identifier

    def __str__(self):
        return f'"Read of {self.__book.title}" by {self.__book.author} read on {self.__date}'

    @property
    def exportable_dictionary(self) -> Dict:
        """ The dictionary"""
        book_dictionary = self.__book.dictionary
        book_dictionary['date'] = self.__date.strftime("%Y-%m-%d")
        book_dictionary['read_number'] = str(self.__read_number)
        return book_dictionary


import os
import json
import csv
from typing import Dict, List
from .printer import Printer
import logging

class Loader:
    """ 
    Utility to load abstract the load from disk. 
    It is able to load CSV (resulting in a `List`) or JSON (resulting in a `Dict`).
    It will detect them by their extension.
    It also offers a printer to write on the originaly loaded file.
    """

    def __init__(self, target_path: str) -> None:
        """ Build a loader for a file with a given FULL path """
        self.__target_path = target_path
        self.__logger = logging.getLogger(__name__)

    @property
    def source_exist(self) -> bool:
        return os.path.exists(self.__target_path)

    @property
    def source_path(self) -> str:
        return self.__target_path

    def printer(self) -> 'Printer':
        return Printer(self.__target_path)

    def load(self):
        """ Load the content of the file pointed by the loader. 
            Warning: This method can return a list or a dictionary."""
        extension = os.path.splitext(self.__target_path)[-1]
        if extension == '.json':
            return self.__load_dictionary()
        elif extension == '.csv':
            return self.__load_array()
        else:
            raise Exception(f"Unknown extension: {extension}")

    def __load_dictionary(self) -> Dict: 
        """ Load the content of the loader target path as a dictionary. """
        self.__logger.info(f"Load file at path: {self.__target_path}")
        with open(self.__target_path, "r", encoding="utf-8") as json_file:
            json_content = json.load(json_file)
            return json_content

    def __load_array(self) -> List:
        """ Use to load from a CVS """
        self.__logger.info(f"Load file at path: {self.__target_path}")
        entries = []
        with open(self.__target_path, "r", encoding="utf-8") as cvs_file:
            file_reader = csv.DictReader(cvs_file)
            for line in file_reader:
                entries.append(line)
            return entries



import logging
from datetime import date

from .bookshelf import Bookshelf
from .printer import OnDirectoryPrinter

READING_OUT_FILE_NAME = "books_reading"
READ_OUT_FILE_NAME_PREFIX = "books_read_"
READ_OUT_NO_DATE_FILE_NAME = "books_read_no_date"
BOOKS_BY_TAGS_FILE_NAME = "books_by_tags"
STATS_FILE_NAME = "books_stats"
FAVOURITES_FILE_NAME = "books_favourites"
TO_READ_FILE_NAME = "books_to_read"

class Generator:
    """ 
    The generato is in charge of writing the resulting files.
    It needs the shelf (`Bookshelf`) and a printer with a fix path (`OnDirectoryPrinter`).
    """

    def __init__(self, bookshelf: Bookshelf, printer: OnDirectoryPrinter) -> None:
        self.__printer = printer
        self.__shelf = bookshelf
        self.__logger = logging.getLogger(__name__)

    def __generate_reading_books(self):
        """ Generate the currently reading books. """
        self.__logger.info('Generating reading books.')
        reading_books = [b.dictionary for b in self.__shelf.reading_books()]
        books_dict = { 'books': reading_books }
        self.__printer.dump(books_dict, READING_OUT_FILE_NAME)

    def __generate_read_books(self):
        """ Generate a file for every year with the read books that year. """
        current_year = date.today().year
        for year in range(2022, 2000, -1):
            books_this_year = [b.exportable_dictionary for b in self.__shelf.reads_on_year(year)]
            if len(books_this_year) > 0:
                stats = self.__shelf.statistics_for_year(year)
                self.__logger.info(f'Generate for year {year}. A total of {len(books_this_year)} books')
                books_dict = { 'stats': stats, 'books': books_this_year }
                self.__printer.dump(books_dict, READ_OUT_FILE_NAME_PREFIX + str(year))

    def __generate_no_date_books(self):
        """ Generate a file with the no date books. """
        read_books_without_date = [b.dictionary for b in self.__shelf.read_on_unkown_date()]
        self.__logger.info(f'Generate for books without read date: {len(read_books_without_date)} books')
        books_dict = { 'books': read_books_without_date }
        self.__printer.dump(books_dict, READ_OUT_NO_DATE_FILE_NAME)

    def __generate_books_by_tags(self):
        """ Generate a file with separated by tags. """
        books_by_tag = self.__shelf.books_by_tags
        self.__logger.info(f'Generate for books by tag, for tags: {", ".join(self.__shelf.tags)}')
        books_dict = { 'books': books_by_tag }
        self.__printer.dump(books_dict, BOOKS_BY_TAGS_FILE_NAME)

    def __generate_stats(self):
        """ Generate a file with the genenal statistics. """
        stats = self.__shelf.statistics
        self.__printer.dump(stats, STATS_FILE_NAME)

    def __generate_favourites(self):
        """ Generate the list of favourite books. """
        if self.__shelf.favourites():
            books = [b.dictionary for b in self.__shelf.favourites()]
            books_dict = { 'books': books }
            self.__printer.dump(books_dict, FAVOURITES_FILE_NAME)

    def __generate_to_read(self):
        """ Generate the to read book list. """
        if self.__shelf.to_read():
            books = [b.dictionary for b in self.__shelf.to_read()]
            books_dict = { 'books': books }
            self.__printer.dump(books_dict, TO_READ_FILE_NAME)
    
    def execute(self) -> None:
        """ Execute all exporting """
        self.__generate_reading_books()
        self.__generate_read_books()
        self.__generate_no_date_books()
        self.__generate_books_by_tags()
        self.__generate_stats()
        self.__generate_favourites()
        self.__generate_to_read()
        self.__logger.info(f"Generation done! You will find the resulting files at: \n'{self.__printer.printer_path}'")